import argparse
import hashlib
import pickle
import random
import re
import numpy as np
import lmdb
import torch

from itertools import chain
from collections.abc import Mapping
from humanfriendly import parse_size

# See https://www.drive5.com/usearch/manual/IUPAC_codes.html
from torch.utils.data import IterableDataset

nucleotide_wildcard_mapping = {
    'R': ['A', 'G'],
    'Y': ['C', 'T'],
    'S': ['G', 'C'],
    'W': ['A', 'T'],
    'K': ['G', 'T'],
    'M': ['A', 'C'],
    'B': ['C', 'G', 'T'],
    'D': ['A', 'G', 'T'],
    'H': ['A', 'C', 'T'],
    'V': ['A', 'C', 'G']
}
nucleotide_wildcards = set(nucleotide_wildcard_mapping)
nucleotide_wildcard_re = re.compile('(' + '|'.join(nucleotide_wildcards) + ')')


def identity(x, *args, **kwargs):
    return x


def merge_dicts(dicts, cat_fn=None):
    """
    Args:
        dicts: Dictionaries to merge key-wise.
        cat_fn (optional): Function to use for merging dictionary values.

    Example:
    >>> merge_dicts({'a': 0, 'b': 42}, {'a': 1})
    {'a': [0, 1], 'b': [42]}
    """
    if cat_fn is None:
        cat_fn = identity

    keys = set(chain.from_iterable(dicts))
    merged = {k: cat_fn([d[k] for d in dicts if k in d]) for k in keys}
    return merged


def digest_strings(ss):
    m = hashlib.md5()
    for s in ss:
        m.update(s.encode())
    return m.hexdigest()


def get_lmdb_keys(env):
    with env.begin() as txn:
        return list(txn.cursor().iternext(values=False))


def to_device(o, device=None):
    if isinstance(o, torch.Tensor):
        return o.to(device)
    if isinstance(o, (list, tuple)):
        return type(o)(to_device(el, device) for el in o)
    if isinstance(o, Mapping):
        return {k: to_device(v, device) for k, v in o.items()}
    return o


def sub_nucleotide_wildcards(seq):
    res = seq
    for match in nucleotide_wildcard_re.finditer(seq):
        res = res[:match.start()] + random.choice(nucleotide_wildcard_mapping[match.group()]) + res[match.end():]
    return res


def read_from_lmdb(path):
    with lmdb.open(str(path), readonly=True) as env:
        with env.begin() as txn, txn.cursor() as cursor:
            for key, value in cursor.iternext():
                yield key.decode(), pickle.loads(value)


return_formats = {
    'torch': (torch.Tensor, torch.tensor),
    'numpy': (np.ndarray, np.array)
}

def to_return_format(x, return_format):
    format = return_formats.get(return_format)

    if format is None:
        raise ValueError(f'Unknown return format {return_format}. '
                         f'Supported formats are {list(return_formats)}')
    
    cls, cast_fn = format

    if isinstance(x, cls):
        return x

    return cast_fn(x)


class IteratorWrapper(IterableDataset):
    def __init__(self, it):
        self.it = it

    def __iter__(self):
        yield from self.it


class HumanFriendlyParsingAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if isinstance(values, list):
            values = [parse_size(v) for v in values]
        else:
            values = parse_size(values)

        setattr(namespace, self.dest, values)


class NestedNamespace(argparse.Namespace):
    def __init__(self, **kwargs):
        super().__init__()
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __setattr__(self, key, value):
        parent, *children = key.split('.', maxsplit=1)
        if children:
            existing = {}
            if hasattr(self, parent):
                existing = vars(getattr(self, parent))
            super().__setattr__(parent, NestedNamespace(**{**existing, children[0]: value}))
        else:
            super().__setattr__(key, value)



import torch
import protenc.utils as utils

from functools import cached_property
from protenc.types import BatchSize, ProteinEncoderInput, ReturnFormat
from torch.utils.data import DataLoader
from protenc.models import BaseProteinEmbeddingModel, get_model


class ProteinEncoder:
    def __init__(
        self,
        model: BaseProteinEmbeddingModel,
        batch_size: BatchSize = None,
        preprocess_workers: int = 0
    ):
        self.model = model
        self.batch_size = 1 if batch_size is None else batch_size
        self.preprocess_workers = preprocess_workers


    @cached_property
    def device(self):
        return next(iter(self.model.parameters())).device
    
    def _get_data_loader(self, proteins: list[str]):
        assert isinstance(self.batch_size, int), 'batch size must be provided as integer at the moment'
        return DataLoader(
            proteins,
            collate_fn=self.model.prepare_sequences,
            batch_size=self.batch_size,
            num_workers=self.preprocess_workers
        )
    
    def _encode(
        self,
        proteins: list[str],
        average_sequence: bool = False,
        return_format: ReturnFormat = 'torch'
    ):
        batches = self._get_data_loader(proteins)

        for batch in batches:
            batch = utils.to_device(batch, self.device)
            
            for embed in self.model(batch):
                if average_sequence:
                    embed = embed.mean(0)
                
                yield utils.to_return_format(embed.cpu(), return_format)

    def encode(
        self,
        proteins: ProteinEncoderInput,
        average_sequence: bool = False,
        autocast: bool = False,
        return_format: ReturnFormat = 'torch'
    ):
        with torch.autocast('cuda', enabled=autocast):
            if isinstance(proteins, dict):
                yield from zip(
                    proteins.keys(),
                    self.encode(
                        list(proteins.values()),
                        average_sequence=average_sequence,
                        return_format=return_format
                    )
                )
            elif isinstance(proteins, list):
                yield from self._encode(
                    proteins,
                    average_sequence=average_sequence,
                    return_format=return_format
                )
            else:
                raise TypeError('Expected list of proteins sequences or dictionary with protein '
                                f'sequences as values but found {type(proteins)}')
    
    def __call__(self, *args, **kwargs):
        return self.encode(*args, **kwargs)


def get_encoder(model_name, device=None, **kwargs):
    model = get_model(model_name, device)
    return ProteinEncoder(model, **kwargs)



import argparse
import json
import pickle
import numpy as np
import lmdb

from Bio import SeqIO
from csv import DictReader
from pathlib import Path
from abc import ABC, abstractmethod
from typing import Callable
from protenc.utils import HumanFriendlyParsingAction


class BaseInputReader(ABC):
    @staticmethod
    def add_arguments_to_parser(parser):
        pass

    @classmethod
    @abstractmethod
    def from_args(cls, path, args):
        pass

    @abstractmethod
    def __iter__(self):
        pass



class BaseOutputWriter:
    @staticmethod
    def add_arguments_to_parser(parser):
        pass

    @classmethod
    @abstractmethod
    def from_args(cls, path, args):
        pass

    @abstractmethod
    def __enter__(self) -> Callable[[str, np.ndarray], None]:
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError


class CSVReader(BaseInputReader):
    def __init__(self,
                 path,
                 delimiter=',',
                 label_col='label',
                 sequence_col='protein'):
        self.path = Path(path)

        self.delimiter = delimiter
        self.label_col = label_col
        self.sequence_col = sequence_col

    @staticmethod
    def add_arguments_to_parser(parser: argparse.ArgumentParser):
        parser.add_argument('--csv_reader.delimiter', default=',')
        parser.add_argument('--csv_reader.label_col', default='label')
        parser.add_argument('--csv_reader.sequence_col', default='protein')

    @classmethod
    def from_args(cls, path, args):
        return cls(
            path,
            delimiter=args.csv_reader.delimiter,
            label_col=args.csv_reader.label_col,
            sequence_col=args.csv_reader.sequence_col
        )

    def __iter__(self):
        with self.path.open() as fp:
            reader = DictReader(fp, delimiter=self.delimiter)
            for row in reader:
                yield row[self.label_col], row[self.sequence_col]



class JSONReader(BaseInputReader):
    def __init__(self, path, stream: bool = False):
        self.path = Path(path)
        self.stream = stream

    @staticmethod
    def add_arguments_to_parser(parser: argparse.ArgumentParser):
        parser.add_argument('--json_reader.stream', action='store_true')

    @classmethod
    def from_args(cls, path, args):
        return cls(path, args.json_reader.stream)

    def __iter__(self):
        if self.stream:
            try:
                import json_stream
            except ImportError:
                raise ImportError('json_stream needs to be installed for streaming json input.')

            json_load = json_stream.load
        else:
            json_load = json.load

        with self.path.open() as fp:
            for label, protein in json_load(fp):
                yield label, protein


class FASTAReader(BaseInputReader):
    def __init__(self, path):
        self.path = Path(path)

    @classmethod
    def from_args(cls, path, args):
        return cls(path)

    def __iter__(self):
        with self.path.open() as fp:
            fasta_sequences = SeqIO.parse(fp, 'fasta')
            for fasta in fasta_sequences:
                label, sequence = fasta.id, str(fasta.seq)
                yield label, sequence



class LMDBWriter(BaseOutputWriter):
    def __init__(self, path, **lmdb_kwargs):
        self.path = path
        self.lmdb_kwargs = lmdb_kwargs
        self.ctx = None

    @staticmethod
    def add_arguments_to_parser(parser: argparse.ArgumentParser):
        parser.add_argument('--lmdb_writer.map_size', action=HumanFriendlyParsingAction, default=2 ** 30)
        parser.add_argument('--lmdb_writer.no_sub_dir', action='store_true')

    @classmethod
    def from_args(cls, path, args):
        return cls(
            path,
            map_size=args.lmdb_writer.map_size,
            subdir=not args.lmdb_writer.no_sub_dir
        )

    def _callback(self, label, embedding):
        assert self.ctx is not None
        assert isinstance(label, str) and isinstance(embedding, np.ndarray)

        _, txn = self.ctx
        txn.put(label.encode(), pickle.dumps(embedding))

    def __enter__(self) -> Callable[[str, np.ndarray], None]:
        if self.ctx is None:
            env = lmdb.open(str(self.path), **self.lmdb_kwargs)
            txn = env.begin(write=True)
            self.ctx = (env, txn)

        return self._callback

    def __exit__(self, exc_type, exc_val, exc_tb):
        env, txn = self.ctx
        txn.commit()
        env.close()


input_format_mapping = {
    'csv': CSVReader,
    'json': JSONReader,
    'fasta': FASTAReader
}

output_format_mapping = {
    'lmdb': LMDBWriter
}



import argparse
import os
import textwrap
import torch
import torch.nn as nn
import protenc

from functools import partial
from pathlib import Path
from torch.utils.data import DataLoader
from tqdm import tqdm
from protenc import io as io, utils
from protenc.models import EmbeddingType


def get_input_reader_cls(args):
    input_path = Path(args.input_path)
    infer = args.input_format in [None, 'infer']
    if infer:
        input_format = input_path.suffix[1:]
    else:
        input_format = args.input_format

    cls = io.input_format_mapping.get(input_format)

    if cls is None:
        raise ValueError(f'Unknown input format \'{input_format}\'' + ' (inferred)' if infer else '')

    return cls


def get_output_reader_cls(args):
    output_path = Path(args.output_path)
    infer = args.input_format in [None, 'infer']
    if infer:
        output_format = output_path.suffix[1:]
    else:
        output_format = args.output_format

    cls = io.output_format_mapping.get(output_format)

    if cls is None:
        raise ValueError(f'Unknown input format \'{output_format}\'' + ' (inferred)' if infer else '')

    return cls


def collate_fn(samples, max_len=None, transform_fn=None):
    labels, sequences = zip(*samples)

    if max_len:
        sequences = [s[:max_len] for s in sequences]

    if transform_fn is not None:
        sequences = transform_fn(sequences)

    return labels, sequences


@torch.no_grad()
def main(args):
    model = protenc.get_model(args.model_name)
    model.eval()

    print(f'Reading data from {args.input_path}')

    input_reader = args.input_reader_cls.from_args(args.input_path, args)

    transform_fn = model.prepare_sequences
    if args.substitute_wildcards:
        transform_fn = lambda seqs: transform_fn([utils.sub_nucleotide_wildcards(s) for s in seqs])

    batches = DataLoader(
        utils.IteratorWrapper(input_reader),
        batch_size=args.batch_size,
        collate_fn=partial(collate_fn, max_len=args.max_prot_len, transform_fn=transform_fn),
        num_workers=args.num_workers
    )

    if 'cuda' in args.device:
        if args.data_parallel:
            model.model = nn.DataParallel(model.model, device_ids=args.device_ids or None)

    model = model.to(args.device)

    print(f'Starting inference loop; writing to {args.output_path}')

    with (
        torch.cuda.amp.autocast(enabled=args.amp),
        args.output_writer_cls.from_args(args.output_path, args) as writer,
        tqdm() as pbar
    ):
        for labels, sequences in batches:
            sequences = utils.to_device(sequences, args.device)
            output = model(sequences)

            for label, embedding in zip(labels, output):
                embedding = embedding.cpu().numpy()

                if args.compute_mean and model.embedding_kind == EmbeddingType.PER_RESIDUE:
                    assert embedding.ndim == 2
                    embedding = embedding.mean(0)

                if args.cast_to:
                    embedding = embedding.astype(args.cast_to)

                if not args.dry_run:
                    # TODO: having the put call on the same thread/process as the GPU calls probably results in stalls.
                    #   This may be improved by setting lmdb.open(...) options appropriately or moving the storing
                    #   procedure into its own thread or process.
                    writer(label, embedding)

            pbar.update(len(labels))


def entrypoint():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=textwrap.dedent('''
            Script to bulk generate protein embeddings for given protein sequences, supporting various pre-trained
            protein language models, a variety of data formats and multi-GPU inference.
            ''')
    )
    parser.add_argument('input_path',
                        help='Path to file containing protein sequence. '
                             f'Currently supported formats are {list(io.input_format_mapping)}. '
                             f'File format is inferred from file extension if not set explicitly '
                             f'with the --input_format option.')
    parser.add_argument('output_path',
                        help='Path to output where embeddings should be written.  '
                             f'Currently supported formats are {list(io.output_format_mapping)}. '
                             f'File format is inferred from file extension if not set explicitly '
                             f'with the --output_format option.')
    parser.add_argument('--input_format', default='infer', choices=['infer', 'fasta', 'csv', 'json'],
                        help=f'Data format of input. Supported formats are {list(io.input_format_mapping)}. '
                             f'Will be inferred from input path by default.')
    parser.add_argument('--output_format', default='infer', choices=['infer', 'parquet', 'pickle', 'lmdb'],
                        help=f'Data format of output. Supported formats are {list(io.output_format_mapping)}. '
                             f'Will be inferred from output path by default.')
    parser.add_argument('--batch_size', type=int, default=64,
                        help='Batch size. Note, that a batch will be distributed to multiple GPUs if the '
                             '--data_parallel flag is set. The value should be adjusted accordingly.')
    parser.add_argument('--model_name', default='prot_bert', choices=protenc.list_models(),
                        help='Name / ID of the embedding model to be loaded.')
    parser.add_argument('--data_parallel', action='store_true',
                        help='Use multiple GPUs with torch.nn.DataParallel.')
    parser.add_argument('--amp', '--fp16', dest='amp', action='store_true',
                        help='Automatic mixed precision aka FP16. Can reduce memory consumption '
                             'and speed up inference if supported.')
    parser.add_argument('--device', default='cuda')
    parser.add_argument('--device_ids', type=int, nargs='*',
                        help='Which devices to use for data parallelism.')
    parser.add_argument('--no_gpu', action='store_true',
                        help='Use the CPU for inference.')
    parser.add_argument('--compute_mean', '--pool', action='store_true',
                        help='Compute the average over the sequence axis of embeddings. '
                             'Applies only if the used model does produce sequence-wise outputs.')
    parser.add_argument('--substitute_wildcards', action='store_true',
                        help='Substitute amino acid wildcards by possible substitutes.')
    parser.add_argument('--cast_to',
                        help='Cast embedding arrays to a numpy data type.')
    parser.add_argument('--max_prot_len', type=int, default=512,
                        help='Maximum length of protein sequences. '
                             'Note: this should always be set to some value as length outliers '
                             'may produce high memory peaks for attention-based models.')
    parser.add_argument('--num_workers', type=int, default=0,
                        help='Number of parallel workers for preprocessing and batching. '
                             'Set to 0 for using the main process only. '
                             'Defaults to zero.')
    parser.add_argument('--dry_run', action='store_true',
                        help='Perform a dry run, i.e. do not write results.')

    args, rem_args = parser.parse_known_args(namespace=utils.NestedNamespace())

    args.input_reader_cls = get_input_reader_cls(args)
    args.input_reader_cls.add_arguments_to_parser(parser)

    args.output_writer_cls = get_output_reader_cls(args)
    args.output_writer_cls.add_arguments_to_parser(parser)

    parser.parse_args(namespace=args)

    main(args)


if __name__ == '__main__':
    entrypoint()



from dataclasses import dataclass
import torch
import torch.nn as nn

from collections import OrderedDict
from typing import Callable
from enum import Enum
from transformers import (
    BertModel,
    BertTokenizer,
    T5EncoderModel,
    T5Tokenizer
)


class EmbeddingType(Enum):
    PER_RESIDUE = 'per_residue'
    PER_PROTEIN = 'per_protein'


class BaseProteinEmbeddingModel(nn.Module):
    embedding_type: EmbeddingType

    def prepare_sequences(self, sequences):
        return NotImplementedError

    def forward(self, input):
        raise NotImplementedError


def load_huggingface_language_model(model_cls, tokenizer_cls, model_name, load_weights=True):
    if load_weights:
        model = model_cls.from_pretrained(model_name)
        tokenizer = tokenizer_cls.from_pretrained(model_name)
        return model, tokenizer
    else:
        config = model_cls.config_class.from_pretrained(model_name)
        model = model_cls(config)
        tokenizer = tokenizer_cls.from_pretrained(model_name)
        return model, tokenizer


class BaseProtTransEmbeddingModel(BaseProteinEmbeddingModel):
    embedding_kind = EmbeddingType.PER_RESIDUE
    available_models = None

    def __init__(self, model, tokenizer):
        super().__init__()
        self.model = model
        self.model.eval()
        self.tokenizer = tokenizer

    def _validate_model_name(self, model_name):
        assert self.available_models is None or model_name in self.available_models, \
            f'Unknown model name \'{model_name}\'. Available options are {self.available_models}'

    def prepare_sequences(self, sequences):
        # ProtTrans tokenizers expect whitespaces between residues
        sequences = [' '.join(s.replace(' ', '')) for s in sequences]

        return self.tokenizer.batch_encode_plus(
            sequences,
            return_tensors='pt',
            add_special_tokens=True,
            padding=True
        )

    def _post_process_embedding(self, embed, seq_len):
        raise NotImplementedError

    @torch.no_grad()
    def forward(self,  input):
        attn_mask = input['attention_mask']

        output = self.model(**input)

        embeddings = output.last_hidden_state.cpu()
        seq_lens = (attn_mask == 1).sum(-1)

        for embed, seq_len in zip(embeddings, seq_lens):
            # Tokenized sequences have the following form:
            # [CLS] V N ... I K [SEP] [PAD] ... [PAD]
            #
            # We remove the special tokens ([CLS], [SEP], [PAD]) before
            # computing the mean over the remaining sequence

            yield self._post_process_embedding(embed, seq_len)


class ProtBERTEmbeddingModel(BaseProtTransEmbeddingModel):
    available_models = [
        'prot_bert',
        'prot_bert_bfd'
    ]

    def __init__(self, model_name: str, load_weights: bool = False):
        self._validate_model_name(model_name)

        mode_name = f'Rostlab/{model_name}'
        model, tokenizer = load_huggingface_language_model(
            BertModel,
            BertTokenizer,
            mode_name,
            load_weights=load_weights
        )

        super().__init__(model=model, tokenizer=tokenizer)

    def _post_process_embedding(self, embed, seq_len):
        return embed[1:seq_len - 1]


class ProtT5EmbeddingModel(BaseProtTransEmbeddingModel):
    available_models = [
        'prot_t5_xl_uniref50',
        'prot_t5_xl_bfd',
        'prot_t5_xxl_uniref50',
        'prot_t5_xxl_bfd'
    ]

    def __init__(self, model_name: str, load_weights: bool = False):
        self._validate_model_name(model_name)

        mode_name = f'Rostlab/{model_name}'
        model, tokenizer = load_huggingface_language_model(
            T5EncoderModel,
            T5Tokenizer,
            mode_name,
            load_weights=load_weights
        )

        super().__init__(model=model, tokenizer=tokenizer)

    def _post_process_embedding(self, embedding, seq_len):
        return embedding[:seq_len - 1]


class ESMEmbeddingModel(BaseProteinEmbeddingModel):
    embedding_kind = EmbeddingType.PER_RESIDUE

    def __init__(self, model_name: str, repr_layer: int):
        super().__init__()

        self.model, self.alphabet = torch.hub.load('facebookresearch/esm:main', model_name)
        self.model.eval()
        self.batch_converter = self.alphabet.get_batch_converter()
        self.repr_layer = repr_layer

    def prepare_sequences(self, sequences):
        _, _, batch_tokens = self.batch_converter([('', seq) for seq in sequences])

        return {
            'tokens': batch_tokens
        }

    @torch.no_grad()
    def forward(self, input):
        results = self.model(**input, repr_layers=[self.repr_layer], return_contacts=True)
        token_representations = results['representations'][self.repr_layer]
        seq_lengths = (input['tokens'] != self.alphabet.padding_idx).sum(1)

        for i, seq_len in enumerate(seq_lengths):
            yield token_representations[i, 1:seq_len - 1]


@dataclass
class ModelDescription:
    name: str
    family: str
    embed_dim: int
    init_fn: Callable[[], BaseProteinEmbeddingModel]


model_descriptions = [
    # ProtTrans family (https://github.com/agemagician/ProtTrans)
    ModelDescription(
        name='prot_t5_xl_uniref50',
        family='ProtTrans',
        embed_dim=1024,
        init_fn=lambda: ProtT5EmbeddingModel('prot_t5_xl_uniref50')
    ),
    ModelDescription(
        name='prot_t5_xl_bfd',
        family='ProtTrans',
        embed_dim=1024,
        init_fn=lambda: ProtT5EmbeddingModel('prot_t5_xl_bfd')
    ),
    ModelDescription(
        name='prot_t5_xxl_uniref50',
        family='ProtTrans',
        embed_dim=1024,
        init_fn=lambda: ProtT5EmbeddingModel('prot_t5_xxl_uniref50')
    ),
    ModelDescription(
        name='prot_t5_xxl_bfd',
        family='ProtTrans',
        embed_dim=1024,
        init_fn=lambda: ProtT5EmbeddingModel('prot_t5_xxl_bfd')
    ),
    ModelDescription(
        name='prot_bert_bfd',
        family='ProtTrans',
        embed_dim=1024,
        init_fn=lambda: ProtBERTEmbeddingModel('prot_bert_bfd')
    ),
    ModelDescription(
        name='prot_bert',
        family='ProtTrans',
        embed_dim=1024,
        init_fn=lambda: ProtBERTEmbeddingModel('prot_bert')
    ),

    # ESM family (https://github.com/facebookresearch/esm)
    ModelDescription(
        name='esm2_t48_15B_UR50D',
        family='ESM',
        embed_dim=5120,
        init_fn=lambda: ESMEmbeddingModel('esm2_t48_15B_UR50D', repr_layer=48)
    ),
    ModelDescription(
        name='esm2_t36_3B_UR50D',
        family='ESM',
        embed_dim=2560,
        init_fn=lambda: ESMEmbeddingModel('esm2_t33_650M_UR50D', repr_layer=33)
    ),
    ModelDescription(
        name='esm2_t33_650M_UR50D',
        family='ESM',
        embed_dim=1280,
        init_fn=lambda: ESMEmbeddingModel('esm2_t48_15B_UR50D', repr_layer=48)
    ),
    ModelDescription(
        name='esm2_t30_150M_UR50D',
        family='ESM',
        embed_dim=640,
        init_fn=lambda: ESMEmbeddingModel('esm2_t30_150M_UR50D', repr_layer=30)
    ),
    ModelDescription(
        name='esm2_t12_35M_UR50D',
        family='ESM',
        embed_dim=480,
        init_fn=lambda: ESMEmbeddingModel('esm2_t12_35M_UR50D', repr_layer=12)
    ),
    ModelDescription(
        name='esm2_t6_8M_UR50D',
        family='ESM',
        embed_dim=480,
        init_fn=lambda: ESMEmbeddingModel('esm2_t6_8M_UR50D', repr_layer=6)
    ),
]


model_dict: dict[str, ModelDescription] = OrderedDict(
    (m.name, m) for m in model_descriptions
)

model_families = set(m.family for m in model_descriptions)


def list_models(family: str | None = None):
    if family is not None:
        if family not in model_families:
            raise ValueError(f'Unknown model family \'{family}\'. Available families are {model_families}')
        
        return [m.name for m in model_descriptions if m.family == family]
    else:
        return list(model_dict)
    

def get_model_info(model_name: str):
    if model_name not in model_dict:
        raise ValueError(f'Unknown model \'{model_name}\'. Available models are {list_models()}')

    model_desc = model_dict[model_name]

    return {
        'name': model_desc.name,
        'family': model_desc.family,
        'embed_dim': model_desc.embed_dim
    }


def get_model(model_name, device=None):
    model = model_dict[model_name].init_fn()

    if device is not None:
        model = model.to(device)
    
    return model
